@@ -33,18 +33,11 @@ extern "C" {
 #include <groonga.h>
 #include "mrn_mysql_compat.h"
 
-#if (MYSQL_VERSION_ID >= 50603) || \
-    (MYSQL_VERSION_ID >= 50513 && MYSQL_VERSION_ID < 50600) || \
-    (MYSQL_VERSION_ID >= 50158 && MYSQL_VERSION_ID < 50500)
-#  define MRN_HANDLER_CLONE_NEED_NAME 1
-#endif
-
 #if (MYSQL_VERSION_ID >= 50514 && MYSQL_VERSION_ID < 50600)
 #  define MRN_HANDLER_HAVE_FINAL_ADD_INDEX 1
 #endif
 
-#if (MYSQL_VERSION_ID >= 50603) || \
-    (defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50209)
+#if (MYSQL_VERSION_ID >= 50603) || defined(MRN_MARIADB_P)
 #  define MRN_HANDLER_HAVE_HA_RND_NEXT 1
 #  define MRN_HANDLER_HAVE_HA_RND_POS 1
 #  define MRN_HANDLER_HAVE_HA_INDEX_READ_MAP 1
@@ -56,8 +49,7 @@ extern "C" {
 #  define MRN_HANDLER_HAVE_HA_INDEX_NEXT_SAME 1
 #endif
 
-#if (MYSQL_VERSION_ID >= 50604) || \
-    (defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50302)
+#if (MYSQL_VERSION_ID >= 50604) || defined(MRN_MARIADB_P)
 #  define MRN_HANDLER_HAVE_HA_CLOSE 1
 #  define MRN_HANDLER_HAVE_MULTI_RANGE_READ 1
 #endif
@@ -77,15 +69,10 @@ extern "C" {
 #  endif
 #endif
 
-#if (defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50302)
+#ifdef MRN_MARIADB_P
 #  define MRN_HANDLER_HAVE_MULTI_RANGE_READ_INFO_KEY_PARTS
 #endif
 
-#if MYSQL_VERSION_ID >= 50500
-#  define MRN_HANDLER_HAVE_TRUNCATE
-#  define MRN_HANDLER_HAVE_GET_PARENT_FOREIGN_KEY_LIST
-#endif
-
 #if MYSQL_VERSION_ID < 50600
 #  define MRN_HANDLER_HAVE_GET_TABLESPACE_NAME
 #endif
@@ -94,10 +81,6 @@ extern "C" {
 #  define MRN_HANDLER_HAVE_SET_HA_SHARE_REF
 #endif
 
-#if MYSQL_VERSION_ID >= 50500
-#  define MRN_TABLE_LIST_INIT_REQUIRE_ALIAS
-#endif
-
 #if MYSQL_VERSION_ID >= 50706 && !defined(MRN_MARIADB_P)
 #  define MRN_BIG_TABLES
 #elif defined(BIG_TABLES)
@@ -110,17 +93,10 @@ extern "C" {
 #  define MRN_HA_ROWS_FORMAT "lu"
 #endif
 
-#if (MYSQL_VERSION_ID < 50519) || \
-    defined(MRN_MARIADB_P) || \
-    (50600 <= MYSQL_VERSION_ID && MYSQL_VERSION_ID < 50604)
+#ifdef MRN_MARIADB_P
 #  define MRN_NEED_FREE_STRING_MEMALLOC_PLUGIN_VAR
 #endif
 
-#if MYSQL_VERSION_ID >= 50500
-#  define MRN_HAVE_HA_EXTRA_ADD_CHILDREN_LIST
-#  define MRN_HAVE_HA_EXTRA_IS_ATTACHED_CHILDREN
-#endif
-
 #ifdef MRN_MARIADB_P
 #  define MRN_HAVE_HA_EXTRA_DETACH_CHILD
 #  define MRN_HAVE_HA_EXTRA_PREPARE_FOR_FORCED_CLOSE
@@ -147,10 +123,6 @@ extern "C" {
 #  define MRN_FIELD_STORE_TIME_NEED_TYPE
 #endif
 
-#if MYSQL_VERSION_ID < 50500
-#  define MRN_HAVE_TL_WRITE_ALLOW_READ
-#endif
-
 #if MYSQL_VERSION_ID < 50706 || defined(MRN_MARIADB_P)
 #  define MRN_HAVE_TL_WRITE_DELAYED
 #endif
@@ -159,9 +131,7 @@ extern "C" {
 #  define MRN_HAVE_TL_WRITE_CONCURRENT_DEFAULT
 #endif
 
-#if (defined(MRN_MARIADB_P) && \
-     ((MYSQL_VERSION_ID >= 50306 && MYSQL_VERSION_ID < 50500) || \
-      MYSQL_VERSION_ID >= 50523))
+#ifdef MRN_MARIADB_P
 #  define MRN_HANDLER_AUTO_REPAIR_HAVE_ERROR
 #endif
 
@@ -259,6 +229,22 @@ struct st_mrn_ft_info
   ha_mroonga *mroonga;
 };
 
+#ifdef MRN_SUPPORT_CUSTOM_OPTIONS
+struct ha_field_option_struct
+{
+  const char *groonga_type;
+  const char *flags;
+};
+
+struct ha_index_option_struct
+{
+  const char *tokenizer;
+  const char *normalizer;
+  const char *token_filters;
+  const char *flags;
+};
+#endif
+
 /* handler class */
 class ha_mroonga: public handler
 {
@@ -433,11 +419,6 @@ public:
 #endif
   int index_next_same(uchar *buf, const uchar *key, uint keylen);
 
-  int read_range_first(const key_range *start_key,
-                       const key_range *end_key,
-                       bool eq_range, bool sorted);
-  int read_range_next();
-
   int ft_init();
   FT_INFO *ft_init_ext(uint flags, uint inx, String *key);
   int ft_read(uchar *buf);
@@ -449,11 +430,7 @@ public:
 
   int reset();
 
-#ifdef MRN_HANDLER_CLONE_NEED_NAME
   handler *clone(const char *name, MEM_ROOT *mem_root);
-#else
-  handler *clone(MEM_ROOT *mem_root);
-#endif
   uint8 table_cache_type();
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ
   ha_rows multi_range_read_info_const(uint keyno, RANGE_SEQ_IF *seq,
@@ -484,9 +461,7 @@ public:
 #endif
   int end_bulk_insert();
   int delete_all_rows();
-#ifdef MRN_HANDLER_HAVE_TRUNCATE
   int truncate();
-#endif // MRN_HANDLER_HAVE_TRUNCATE
   double scan_time();
   double read_time(uint index, uint ranges, ha_rows rows);
   const key_map *keys_to_use_for_scanning();
@@ -570,9 +545,7 @@ protected:
 #endif
   bool can_switch_engines();
   int get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
-#ifdef MRN_HANDLER_HAVE_GET_PARENT_FOREIGN_KEY_LIST
   int get_parent_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
-#endif
   uint referenced_by_foreign_key();
   void init_table_handle_for_HANDLER();
   void free_foreign_key_create_info(char* str);
@@ -609,14 +582,23 @@ private:
   void clear_search_result();
   void clear_search_result_geo();
   void clear_indexes();
-  int alter_share_add(const char *path, TABLE_SHARE *table_share);
+  int add_wrap_hton(const char *path, handlerton *wrap_handlerton);
   void remove_related_files(const char *base_path);
   void remove_grn_obj_force(const char *name);
   int drop_index(MRN_SHARE *target_share, uint key_index);
+  int drop_indexes_normal(const char *table_name, grn_obj *table);
+  int drop_indexes_multiple(const char *table_name, grn_obj *table);
+  int drop_indexes(const char *table_name);
+  bool find_column_flags(Field *field, MRN_SHARE *mrn_share, int i,
+                         grn_obj_flags *column_flags);
+  grn_obj *find_column_type(Field *field, MRN_SHARE *mrn_share, int i,
+                            int error_code);
+  grn_obj *find_tokenizer(KEY *key, MRN_SHARE *mrn_share, int i);
   grn_obj *find_tokenizer(const char *name, int name_length);
-  grn_obj *find_normalizer(KEY *key_info);
-  bool find_index_column_flags(KEY *key_info, grn_obj_flags *index_column_flags);
-  bool find_token_filters(KEY *key_info, grn_obj *token_filters);
+  grn_obj *find_normalizer(KEY *key);
+  grn_obj *find_normalizer(KEY *key, const char *name);
+  bool find_index_column_flags(KEY *key, grn_obj_flags *index_column_flags);
+  bool find_token_filters(KEY *key, grn_obj *token_filters);
   bool find_token_filters_put(grn_obj *token_filters,
                               const char *token_filter_name,
                               int token_filter_name_length);
@@ -744,6 +726,15 @@ private:
   int storage_encode_multiple_column_key(KEY *key_info,
                                          const uchar *key, uint key_length,
                                          uchar *buffer, uint *encoded_length);
+  int storage_encode_multiple_column_key_range(KEY *key_info,
+                                               const uchar *start,
+                                               uint start_size,
+                                               const uchar *end,
+                                               uint end_size,
+                                               uchar *min_buffer,
+                                               uint *min_encoded_size,
+                                               uchar *max_buffer,
+                                               uint *max_encoded_size);
   int storage_encode_multiple_column_key_range(KEY *key_info,
                                                const key_range *start,
                                                const key_range *end,
@@ -793,17 +784,15 @@ private:
   int close_databases();
   int ensure_database_open(const char *name);
   int ensure_database_remove(const char *name);
-  int wrapper_delete_table(const char *name, MRN_SHARE *tmp_share,
-                           const char *table_name);
-  int wrapper_delete_index(const char *name, MRN_SHARE *tmp_share,
-                           const char *table_name);
-  int storage_delete_table(const char *name, MRN_SHARE *tmp_share,
+  int wrapper_delete_table(const char *name, handlerton *wrap_handlerton,
                            const char *table_name);
+  int generic_delete_table(const char *name, const char *table_name);
   int wrapper_open(const char *name, int mode, uint test_if_locked);
   int wrapper_open_indexes(const char *name);
   int storage_open(const char *name, int mode, uint test_if_locked);
   int open_table(const char *name);
   int storage_open_columns(void);
+  void storage_close_columns(void);
   int storage_open_indexes(const char *name);
   void wrapper_overwrite_index_bits();
   int wrapper_close();
@@ -930,14 +919,6 @@ private:
   int storage_index_last(uchar *buf);
   int wrapper_index_next_same(uchar *buf, const uchar *key, uint keylen);
   int storage_index_next_same(uchar *buf, const uchar *key, uint keylen);
-  int wrapper_read_range_first(const key_range *start_key,
-                               const key_range *end_key,
-                               bool eq_range, bool sorted);
-  int storage_read_range_first(const key_range *start_key,
-                               const key_range *end_key,
-                               bool eq_range, bool sorted);
-  int wrapper_read_range_next();
-  int storage_read_range_next();
   int generic_ft_init();
   int wrapper_ft_init();
   int storage_ft_init();
@@ -992,13 +973,8 @@ private:
   void storage_cond_pop();
   bool wrapper_get_error_message(int error, String *buf);
   bool storage_get_error_message(int error, String *buf);
-#ifdef MRN_HANDLER_CLONE_NEED_NAME
   handler *wrapper_clone(const char *name, MEM_ROOT *mem_root);
   handler *storage_clone(const char *name, MEM_ROOT *mem_root);
-#else
-  handler *wrapper_clone(MEM_ROOT *mem_root);
-  handler *storage_clone(MEM_ROOT *mem_root);
-#endif
   uint8 wrapper_table_cache_type();
   uint8 storage_table_cache_type();
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ
@@ -1054,9 +1030,7 @@ private:
                               const char *function_name);
   int wrapper_delete_all_rows();
   int storage_delete_all_rows();
-#ifdef MRN_HANDLER_HAVE_TRUNCATE
   int wrapper_truncate();
-#endif // MRN_HANDLER_HAVE_TRUNCATE
   int wrapper_truncate_index();
   int storage_truncate();
   int storage_truncate_index();
@@ -1221,10 +1195,8 @@ private:
   bool storage_can_switch_engines();
   int wrapper_get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
   int storage_get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
-#ifdef MRN_HANDLER_HAVE_GET_PARENT_FOREIGN_KEY_LIST
   int wrapper_get_parent_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
   int storage_get_parent_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
-#endif
   uint wrapper_referenced_by_foreign_key();
   uint storage_referenced_by_foreign_key();
   void wrapper_init_table_handle_for_HANDLER();
